#include <CONFIG.h>
#include <SensorManager.h>
#include <Mail.h>
#include <EventFlags.h>
#include <stm32l4xx.h>
#include <stm32l4xx_hal.h>
#include <LSM9DS0.h>
#include <CommonStructures.h>
#include <MadgwickAHRS.h>
#include <math.h>

SensorManager::SensorManager(EventFlags *sensor_event_flags, Mail<VehicleState, 16> *state_output_mailbox) : lsm(
        LSM9DS0_SDA, LSM9DS0_SCL, LSM9DS0_G_ADDR, LSM9DS0_XM_ADDR) {
    this->sensor_flags = sensor_event_flags;
    this->sensor_tx = state_output_mailbox;
    lsm.calLSM9DS0(lsm.gbias, lsm.abias);
};

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wmissing-noreturn"

void SensorManager::run() {
    while (true) {
        // FIXME: If data is not ready on a tick boundary, the filter will not work
        // Wait for new data, and for a tick (runs at constant frequency)
        sensor_flags->wait_all(0x011);
        sensor_flags->clear(0x011);
        lsm.readAccel();
        lsm.readGyro();
        lsm.readMag();
        // Cache calculated magnetometer and accelerometer values
        // Don't cache gyro values because they have to pass through the filter
        float ax, ay, az;
        float mx, my, mz;
        ax = lsm.calcAccel(lsm.ax) - lsm.abias[0];
        ay = lsm.calcAccel(lsm.ay) - lsm.abias[1];
        az = lsm.calcAccel(lsm.az) - lsm.abias[2];
        mx = lsm.calcAccel(lsm.mx);
        my = lsm.calcAccel(lsm.my);
        mz = lsm.calcAccel(lsm.mz);
        MadgwickAHRSupdate(lsm.calcGyro(lsm.gx) - lsm.gbias[0],
                           lsm.calcGyro(lsm.gy) - lsm.gbias[1],
                           lsm.calcGyro(lsm.gz) - lsm.gbias[2],
                           ax,
                           ay,
                           az,
                           mx,
                           my,
                           mz);
        VehicleState *current_state = sensor_tx->alloc();
        current_state->gps_ok = false;
        current_state->optical_flow_ok = false;
        current_state->tof_ok = false;
        current_state->ax = ax;
        current_state->ay = ay;
        current_state->az = az;
        SensorManager::quaternion_to_tait_bryan(tait_bryan, MadgwickAHRS_h::q0, MadgwickAHRS_h::q1, MadgwickAHRS_h::q2, MadgwickAHRS_h::q3);
        current_state->gz = tait_bryan[0];
        current_state->gy = tait_bryan[1];
        current_state->gx = tait_bryan[2];
        sensor_tx->put(current_state);
    }
};
#pragma clang diagnostic pop

// Outputs a 3-2-1 Tait_Bryan angle set
// FIXME: Check these, also, type of output may have to change to point to the right thing somehow
void SensorManager::quaternion_to_tait_bryan(float* output, float q0, float q1, float q2, float q3) {
    output[0] = atan2((q2 * q3 + q0 * q1), 1.0f / 2.0f - (pow(q1, 2.0f) + pow(q2, 2.0f)));
    output[1] = asin(-2 * (q1 * q3 - q0 * q2));
    output[2] = atan2((q1 * q2 + q0 * q3), 1.0f / 2.0f - (pow(q2, 2.0f) + pow(q3, 2.0f)));
}

void SensorManager::sensor_dready_rising_isr(EventFlags *sensor_events) {
    sensor_events->set(0x01);
}

void SensorManager::sensor_timer_tick_isr(EventFlags *sensor_events) {
    sensor_events->set(0x001);
}